/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "graphwidget.h"
#include<iostream>
#include<cstdio>
#include<math.h>

TGraphArea::TGraphArea() 
{

}
TGraphArea::~TGraphArea()
{
    
}

void TGraphArea::set_master(TGraphWidget * ms)
{
    master = ms;
}

void TGraphArea::on_realize()
{
  // We need to call the base on_realize()
  Gtk::DrawingArea::on_realize();

  // Now we can allocate any additional resources we need
  Glib::RefPtr<Gdk::Window> window = get_window();
  gc = Gdk::GC::create(window);

  Glib::RefPtr<Gdk::Colormap> colormap = get_default_colormap ();

  red_ = Gdk::Color("red");
  black_ = Gdk::Color("black");
  white_ = Gdk::Color("white");
  blue_ = Gdk::Color("blue");

  colormap->alloc_color(red_);
  colormap->alloc_color(black_);
  colormap->alloc_color(white_);
  colormap->alloc_color(blue_);
    
  window->set_background(white_);
}

bool TGraphArea::on_expose_event(GdkEventExpose* e)
{
    // This is where we draw on the window
    Glib::RefPtr<Gdk::Window> window = get_window();
    window->clear();
    master->invalidateGraphState();
    int wWidth = get_width();
    int wHeight = get_height();
    double scale_x = master->get_scale_x();
    double scale_y = master->get_scale_y();
    double grid_x_step = master->get_grid_x_step();
    double grid_y_step = master->get_grid_y_step();
    double axe_x_step = master->get_axe_x_step();
    double axe_y_step = master->get_axe_y_step();
    std::string axe_x_label_format = master->get_axe_x_label_format();
    std::string axe_y_label_format = master->get_axe_y_label_format();
    int inset_x = master->get_inset_x();
    int inset_y = master->get_inset_y();

    // draw the grid
    gc->set_foreground(black_);
    gc->set_line_attributes(1, 
                            Gdk::LINE_ON_OFF_DASH,
                            Gdk::CAP_NOT_LAST,
                            Gdk::JOIN_ROUND);
    double real_pos;
    int linepos;
    // X grid
    if(master->get_grid_x_step()) {
        real_pos = master->grid_x_start_line;
        linepos = inset_x + int(round(
                               (real_pos - master->get_start_x())
                               *scale_x));
        while(linepos <= (wWidth - inset_x)) {
            window->draw_line(gc, 
                              linepos, 
                              inset_y, 
                              linepos, 
                              wHeight - inset_y);
            real_pos+=grid_x_step;
            linepos = inset_x + int(round(
                               (real_pos - master->get_start_x())
                               *scale_x));
        }
    }
    // Y grid
    if(master->get_grid_y_step()) {    
        real_pos = master->grid_y_start_line;
        linepos = wHeight - inset_y - int(round(
                               (real_pos - master->get_start_y())
                               *scale_y));
        while(linepos >= inset_y) {
            window->draw_line(gc, 
                              inset_x,
                              linepos,
                              wWidth - inset_x,
                              linepos);
            real_pos+=grid_y_step;
            linepos = wHeight - inset_y - int(round(
                               (real_pos - master->get_start_y())
                               *scale_y));
        }
    }
    // draw the line for each series
    gc->set_foreground(red_);
    gc->set_line_attributes(2, 
                            Gdk::LINE_SOLID,
                            Gdk::CAP_NOT_LAST,
                            Gdk::JOIN_ROUND);
    for(unsigned sk = 0; sk<master->model->series.size(); sk++) {
        std::vector<Gdk::Point> points_array;
        
        if(master->get_auto_scale_x()&&
           master->get_auto_scale_y()) {
            // the simplest mode
            // all points are within a pane
               for(unsigned i=0; i<master->model->series[sk].points.size();i++) {        
                int xpos = inset_x + int(round(
                           (master->model->series[sk].points[i].x 
                            - master->get_start_x())
                           *scale_x));
                int ypos = get_height() - inset_y -
                           int(round(
                            (master->model->series[sk].points[i].y 
                             - master->get_start_y())
                            *scale_y));
                points_array.push_back(
                   Gdk::Point(xpos,ypos));
            }
        }
        
        // draw in autoscale_x and autoscale_y modes
        window->draw_lines(gc, points_array);
    } // for sk
    
    //Draw some text.
    Glib::RefPtr<Pango::Layout> pangolayout = create_pango_layout(""); 
    char label_buf[10];
    gc->set_foreground(blue_);
    // X grid labels
    if(master->get_grid_x_step()) {
        real_pos = master->grid_x_start_line + axe_x_step;
        linepos = inset_x + int(round(
                               (real_pos - master->get_start_x())
                               *scale_x));
        while(linepos <= (wWidth - inset_x - 20)) {
            std::snprintf(label_buf, 9, axe_x_label_format.c_str(), real_pos);
            pangolayout->set_text(Glib::ustring(label_buf));
            int text_w, text_h;
            pangolayout->get_pixel_size(text_w, text_h);
            window->draw_layout(gc, 
                                linepos + 2,
                                wHeight - inset_y - text_h - 2, 
                                pangolayout);
            real_pos+=axe_x_step;
            linepos = inset_x + int(round(
                               (real_pos - master->get_start_x())
                               *scale_x));
        }
    }
    // Y grid labels
    if(master->get_grid_y_step()) {
        // skip the first entry (below the bottom edge)
        real_pos = master->grid_y_start_line + axe_y_step;
        linepos = wHeight - inset_y - int(round(
                               (real_pos - master->get_start_y())
                               *scale_y));
        while(linepos >= inset_y) {
            std::snprintf(label_buf, 9, axe_x_label_format.c_str(), real_pos);
            pangolayout->set_text(Glib::ustring(label_buf));
            window->draw_layout(gc, 
                                inset_x + 2,
                                linepos + 2, 
                                pangolayout);
            real_pos+=axe_y_step;
            linepos = wHeight - inset_y - int(round(
                               (real_pos - master->get_start_y())
                               *scale_y));
        }
    }
    return true;
}

void TGraphArea::on_size_request(Gtk::Requisition* requisition)
{
  *requisition = Gtk::Requisition();
  // this is the minimal
  requisition->height = 100;
  requisition->width = 100;
}

TGraphWidget::TGraphWidget(TGraphModel * md):
Gtk::Table(2,2),
model(md),
forceInvalidate(true),
auto_scale_x(false),
auto_scale_y(false),
scale_x(1.0),
scale_y(1.0),
start_x(0.0),
start_y(0.0),
grid_x_step(0.0),
grid_y_step(0.0),
axe_x_step(0.0),
axe_y_step(0.0),
axe_x_label_format(""),
axe_y_label_format(""),
inset_x(10),
inset_y(10)
{
    // initialize the child widgets
    hScrollbar = manage(new Gtk::HScrollbar);
    vScrollbar = manage(new Gtk::VScrollbar);
    drawArea = manage(new TGraphArea());
    drawArea->set_master(this);

/*   attach(*hScrollbar, 0, 1, 1, 2, Gtk::FILL,Gtk::SHRINK);    
    attach(*vScrollbar, 1, 2, 0, 1, Gtk::SHRINK,Gtk::FILL);
*/
    attach(*drawArea, 0, 1, 0, 1, Gtk::FILL|Gtk::EXPAND, Gtk::FILL|Gtk::EXPAND);
    
    show_all_children();
}


TGraphWidget::~TGraphWidget()
{}

    
void TGraphWidget::invalidateGraphState() {
    if(forceInvalidate && model->series.size()) {
        // recalculate the X scale and offsets
        if(auto_scale_x) {
            if(model->series[0].points.size()) {
                double minValue = model->series[0].points[0].x;
                double maxValue = minValue;
                for(unsigned i=0; i<model->series[0].points.size();i++) {
                    if(model->series[0].points[i].x>maxValue) {
                        maxValue = model->series[0].points[i].x;
                    } else
                    if(model->series[0].points[i].x<minValue) {
                        minValue = model->series[0].points[i].x;
                    }           
                }
                scale_x = (drawArea->get_width() - 2*inset_x) /
                          (maxValue - minValue);
                start_x = minValue;
            }
        }
        // recalculate the X scale and offsets
        if(auto_scale_y) {
            if(model->series[0].points.size()) {
                double minValue = model->series[0].points[0].y;
                double maxValue = minValue;
                for(unsigned i=0; i<model->series[0].points.size();i++) {
                    if(model->series[0].points[i].y>maxValue) {
                        maxValue = model->series[0].points[i].y;
                    } else
                    if(model->series[0].points[i].y<minValue) {
                        minValue = model->series[0].points[i].y;
                    }           
                }
                scale_y = (drawArea->get_height() - 2*inset_y) / 
                          (maxValue - minValue);
                start_y = minValue;                
            }
        }
        // calculate the starting grid lines coordinates
        if(grid_x_step) {
            grid_x_start_line = (floor(start_x/grid_x_step)) * grid_x_step;
        }
        if(grid_y_step) {
            grid_y_start_line = (floor(start_y/grid_y_step)) * grid_y_step;        
        }

        forceInvalidate = false;        
    } // if model->series.size()        
}
void TGraphWidget::on_size_allocate(Gtk::Allocation& allocation) {
    forceInvalidate = true;
    Gtk::Table::on_size_allocate(allocation);
}
void TGraphWidget::set_auto_scale_x(const bool new_value) {
    // recalculate the scale from the model
    // assuming one series at this time
    auto_scale_x = new_value;
    forceInvalidate = true;
    // just repaint the graph
    queue_draw();
}
void TGraphWidget::set_auto_scale_y(const bool new_value) {
    // recalculate the scale from the model
    auto_scale_y = new_value;
    forceInvalidate = true;
    // just repaint the graph
    queue_draw();
}
void TGraphWidget::set_scale_x(const double new_value) {
    // setting the scale for x manually means
    // the cancellation of the autoscale
    auto_scale_x = false;
    scale_x = new_value;
    // just repaint the graph
    queue_draw();
}
void TGraphWidget::set_scale_y(const double new_value) {
    // setting the scale for y manually means
    // the cancellation of the autoscale
    auto_scale_y = false;
    scale_y = new_value;
    // just repaint the graph
    queue_draw();
}
void TGraphWidget::set_start_x(const double new_value) {
    // setting the start x manually means
    // the cancellation of the autoscale
    auto_scale_x = false;
    start_x = new_value;
    // shift the scrollbar if available
    // ...
    // just repaint the graph
    queue_draw();
}
void TGraphWidget::set_start_y(const double new_value) {
    // setting the start y manually means
    // the cancellation of the autoscale
    auto_scale_y = false;
    start_y = new_value;
    // shift the scrollbar if available
    // ...
    // just repaint the graph
    queue_draw();
}
void TGraphWidget::set_grid_x_step(const double new_value) {
    grid_x_step = new_value;
    forceInvalidate = true;
    queue_draw();
}
void TGraphWidget::set_grid_y_step(const double new_value){
    grid_y_step = new_value;
    forceInvalidate = true;
    queue_draw();
}
void TGraphWidget::set_axe_x_step(const double new_value){
    axe_x_step = new_value;
    forceInvalidate = true;
    queue_draw();
}
void TGraphWidget::set_axe_y_step(const double new_value){
    axe_y_step = new_value;
    forceInvalidate = true;
    queue_draw();
}
void TGraphWidget::set_axe_x_label_format(const std::string new_value){
    axe_x_label_format = new_value;
    forceInvalidate = true;
    queue_draw();
}
void TGraphWidget::set_axe_y_label_format(const std::string new_value){
    axe_y_label_format = new_value;
    forceInvalidate = true;
    queue_draw();
}
void TGraphWidget::set_inset_x(int new_value){
    inset_x = new_value;
    forceInvalidate = true;
    queue_draw();
}
void TGraphWidget::set_inset_y(int new_value){
    inset_y = new_value;
    forceInvalidate = true;
    queue_draw();
}

const bool TGraphWidget::get_auto_scale_x() {
    return auto_scale_x;
}
const bool TGraphWidget::get_auto_scale_y() {
    return auto_scale_y;
}
const double TGraphWidget::get_scale_x() {
    return scale_x;
}
const double TGraphWidget::get_scale_y() {
    return scale_y;
}
const double TGraphWidget::get_start_x() {
    return start_x;
}
const double TGraphWidget::get_start_y() {
    return start_y;
}
const double TGraphWidget::get_grid_x_step(){
    return grid_x_step;
}
const double TGraphWidget::get_grid_y_step(){
    return grid_y_step;
}
const double TGraphWidget::get_axe_x_step(){
    return axe_x_step;
}
const double TGraphWidget::get_axe_y_step(){
    return axe_y_step;
}
const std::string TGraphWidget::get_axe_x_label_format(){
    return axe_x_label_format;
}
const std::string TGraphWidget::get_axe_y_label_format(){
    return axe_y_label_format;
}
const int TGraphWidget::get_inset_x() {
    return inset_x;
}
const int TGraphWidget::get_inset_y() {
    return inset_y;
}
